#include "LtrCrate.h"
#include <QScopedPointer>
#include "lqmeas/devs/ltr/modules/LtrModule.h"
#include "lqmeas/devs/ltr//LtrResolver.h"
#include "lqmeas/devs/DeviceValidator.h"
#include "lqmeas/log/Log.h"
#include "lqmeas/lqtdefs.h"

namespace LQMeas {
    LtrCrate::LtrCrate(TLTR hnd, QString serial, TLTR_CRATE_INFO info) :
        m_serial(serial), m_hnd(hnd) {

        m_iface = info.CrateInterface == LTR_CRATE_IFACE_USB ? IFACE_USB :
                  info.CrateInterface == LTR_CRATE_IFACE_TCPIP ? IFACE_ETH :
                                                                 IFACE_UNKNOWN;

        switch(info.CrateType) {
            case LTR_CRATE_TYPE_LTR010:
                m_typeName = "LTR-U-8/16";
                m_slot_cnt = 16;
                break;
            case LTR_CRATE_TYPE_LTR021:
                m_typeName = "LTR-U-1";
                m_slot_cnt = 1;
                break;
            case LTR_CRATE_TYPE_LTR030:
                m_typeName = "LTR-EU-8/16";
                m_slot_cnt = 16;
                break;
            case LTR_CRATE_TYPE_LTR031:
                m_typeName = "LTR-EU-2";
                m_slot_cnt = 2;
                break;
            case LTR_CRATE_TYPE_LTR032:
                m_typeName = "LTR-E-7/15";
                m_slot_cnt = 16;
                break;
            case LTR_CRATE_TYPE_LTR_CU_1:
                m_typeName = "LTR-CU-1";
                m_slot_cnt = 1;
                break;
            case LTR_CRATE_TYPE_LTR_CEU_1:
                m_typeName = "LTR-CEU-1";
                m_slot_cnt = 1;
                break;
            default:
                m_typeName = "Unknown";
                m_slot_cnt = 16;
                LQMeasLog->report(LogLevel::Warning, "Unknown crate type!", this);
                break;
        }
        m_modules.resize(m_slot_cnt);
    }

    LtrCrate::~LtrCrate() {
        if (isOpened())
            close();
    }

    bool LtrCrate::isOpened() const {
        return LTR_IsOpened(&m_hnd)==LTR_OK;
    }



    quint32 LtrCrate::ltrdIpAddr() {
        return m_hnd.saddr;
    }

    quint16 LtrCrate::ltrdTcpPort() {
        return m_hnd.sport;
    }

    QList<QSharedPointer<LtrModule> > LtrCrate::getModuleList(QStringList nameFilter) {
        QScopedPointer<DeviceValidator> val(new DeviceNameValidator(nameFilter));
        return getModuleList(val.data());
    }



    QList<QSharedPointer<LtrModule> > LtrCrate::getModuleList(const DeviceValidator *validator) {
        QList<QSharedPointer<LtrModule> > ret;
        bool close_req = false;
        if (!isOpened()) {
            if (open().isSuccess()) {
                close_req = true;
            }
        }

        if (isOpened()) {
            WORD mids[LTR_MODULES_PER_CRATE_MAX];
            int err = LTR_GetCrateModules(&m_hnd, mids);
            if (err==LTR_OK) {
                for (int slot=0; (slot < LTR_MODULES_PER_CRATE_MAX) && (slot < m_modules.size()); slot++) {
                    if ((mids[slot]!=LTR_MID_EMPTY) && (mids[slot]!=LTR_MID_IDENTIFYING)) {
                        QSharedPointer<LtrModule> module = m_modules[slot];
                        if (!module || (module->mid()!= mids[slot])) {
                            m_modules[slot] = module = LtrResolver::createModule(
                                        LtrResolver::resolver()->getCrate(m_serial),
                                        slot+LTR_CC_CHNUM_MODULE1, mids[slot]);
                        }

                        if (module && ((validator == 0) || validator->deviceValid(module))) {
                            ret.append(module);
                        }

                    } else {
                        /* если пуст - то удаляем старую информацию о модуле */
                        m_modules[slot] = QSharedPointer<LtrModule>();
                    }
                }
            }


            if (close_req)
                close();
        }
        return ret;
    }

    QList<QSharedPointer<Device> > LtrCrate::getDevList() {
        QList<QSharedPointer<LtrModule> > ltrlist = getModuleList(0);
        QList<QSharedPointer<Device> > retlist;
        foreach (QSharedPointer<LtrModule> ltr, ltrlist) {
            retlist.append(ltr);
        }
        return retlist;
    }

    Error LtrCrate::error(int err) const {
        return err == LTR_OK ? Error::Success() :
                               Error(err, QSTRING_FROM_CSTR(LTR_GetErrorString(err)),
                                     Error::ERRTYPE_DEVNATIVE_LTR);
    }


    Error LtrCrate::protOpen() {
        return error(LTR_Open(&m_hnd));
    }

    Error LtrCrate::protClose() {
        return error(LTR_Close(&m_hnd));
    }

    Error LtrCrate::protLoadConfig(QSettings &set) {
        return error(LTR_ERROR_NOT_IMPLEMENTED);
    }

    Error LtrCrate::protSaveConfig(QSettings &set) const {
        return error(LTR_ERROR_NOT_IMPLEMENTED);
    }

    Error LtrCrate::protConfigure() {
        return Error::Success();
    }
}
