#include "ltr24.h"
#include "lqmeas/lqtdefs.h"
#include "lqmeas/devs/ltr/crates/LtrCrate.h"

namespace LQMeas {

    static double f_ranges[] = {2., 10.};

    static const double f_adc_freqs[] = {
            117.1875e3,  78.125e3,   58.59375e3,   39.0625e3,
            29.296875e3, 19.53125e3, 14.6484375e3, 9.765625e3,
            7.32421875e3, 4.8828125e3, 3.662109375e3, 2.44140625e3,
            1.8310546875e3, 1.220703125e3, 0.91552734375e3, 0.6103515625e3
    };

    QString LTR24::typeModuleName() {
        return "LTR24";
    }


    LTR24::LTR24(QSharedPointer<LtrCrate> crate, unsigned slot, QObject *parent) :
        LtrModule(crate, slot, typeModuleID, parent), DevAdc(this, DevAdc::TypeParallel),
        m_receiver(this) {

        LTR24_Init(&m_hnd);
        adcSetChCnt(LTR24_CHANNEL_NUM);
    }


    bool LTR24:: isOpened() const {
        return LTR24_IsOpened(&m_hnd)==LTR_OK;
    }

    bool LTR24::adcIsRunning() const {
        return m_hnd.Run ? true : false;
    }

    bool LTR24::adcChEnabled(int ch) const {
        Q_ASSERT((ch>=0) && (ch < LTR24_CHANNEL_NUM));
        return m_cfg.Ch[ch].Enabled;
    }

    Error LTR24::adcGetData(double *data, unsigned size, unsigned flags, unsigned tout,
                           unsigned *recvd_size) {
        Error err = Error::Success();
        int wrds_cnt = m_hnd.DataFmt == LTR24_FORMAT_20 ? size : 2*size;
        unsigned recvd_words=0;
        DWORD* wrds;
        QScopedArrayPointer<BOOL> ovrlds(new BOOL[size]);
        INT proc_size = 0;

        if (err.isSuccess()) {
            err = m_receiver.getFrames(wrds_cnt, tout, &wrds, &recvd_words);
        }

        if (err.isSuccess() && (recvd_words!=0)) {
            DWORD proc_flags = LTR24_PROC_FLAG_VOLT | LTR24_PROC_FLAG_AFC_COR | LTR24_PROC_FLAG_CALIBR;
            proc_size = (INT)recvd_words;

            err = error(LTR24_ProcessData(&m_hnd, wrds, data, &proc_size,
                                          proc_flags, ovrlds.data()));
            if (!err.isSuccess()) {
                LQMeasLog->error(tr("Process data error"), err, this);
            } else {
                if (m_hnd.DataFmt == LTR24_FORMAT_24) {
                    for (unsigned ch_idx = 0; ch_idx < adcEnabledChCnt(); ch_idx++) {
                        ChannelStatus status = ChStatusOk;
                        for (unsigned i=ch_idx; ((INT)i < proc_size) && (status == ChStatusOk); i+=adcEnabledChCnt()) {
                            if (ovrlds[i]) {
                                status = ChStatusOverload;
                            }
                        }
                        adcSetChStatus(adcChNum(ch_idx), status);
                    }
                }
            }
        }

        if (err.isSuccess() && recvd_size)
            *recvd_size = proc_size;

        return err;
    }

    TLTR *LTR24::channel() const {
        return &m_hnd.Channel;
    }

    Error LTR24::protOpen() {
        Error err = error(LTR24_Open(&m_hnd, crate()->ltrdIpAddr(), crate()->ltrdTcpPort(),
                                     crate()->serial().toLatin1(), slot()));
        if (err.isSuccess()) {
            err = error(LTR24_GetConfig(&m_hnd));
        }

        if (!err.isSuccess()) {
            LTR24_Close(&m_hnd);
        }
        return err;
    }

    Error LTR24::protClose() {
        return error(LTR24_Close(&m_hnd));
    }

    Error LTR24::protConfigure() {
        m_hnd.ADCFreqCode = m_cfg.ADCFreqCode;
        m_hnd.DataFmt = m_cfg.DataFmt;
        m_hnd.ISrcValue = m_cfg.ISrcValue;
        m_hnd.TestMode = m_cfg.TestMode;
        for (int ch=0; ch < LTR24_CHANNEL_NUM; ch++) {
            m_hnd.ChannelMode[ch].Enable = m_cfg.Ch[ch].Enabled;
            m_hnd.ChannelMode[ch].AC = m_cfg.Ch[ch].AC;
            m_hnd.ChannelMode[ch].Range = m_cfg.Ch[ch].Range;
            m_hnd.ChannelMode[ch].ICPMode = m_cfg.Ch[ch].ICPMode;
        }

        return error(LTR24_SetADC(&m_hnd));
    }

    Error LTR24::protLoadConfig(QSettings &set) {
        int intval;
        bool ok;
        AdcConfig cfg;

        memset(&cfg, 0, sizeof(cfg));

        unsigned ch_cnt = set.beginReadArray("Channels");
        if (ch_cnt == 0) {
            ch_cnt = 1;
            cfg.Ch[0].Enabled = true;
            cfg.Ch[0].Range = LTR24_RANGE_10;
            cfg.Ch[0].ICPMode = FALSE;
            cfg.Ch[0].AC = FALSE;
        } else {
            for (unsigned ch=0; (ch < ch_cnt) && (ch< adcChannelsCnt()); ch++) {
                set.setArrayIndex(ch);
                cfg.Ch[ch].Enabled = set.value("Enabled").toBool();
                cfg.Ch[ch].Range   = set.value("Range").toInt();
                cfg.Ch[ch].ICPMode = set.value("ICPMode").toBool();
                cfg.Ch[ch].AC      = set.value("AC").toBool();
            }
        }

        for (unsigned ch=ch_cnt; ch < adcChannelsCnt(); ch++) {
            cfg.Ch[ch].Enabled = false;
            cfg.Ch[0].Range = LTR24_RANGE_10;
            cfg.Ch[0].ICPMode = FALSE;
            cfg.Ch[0].AC = FALSE;
        }
        set.endArray();

        intval = set.value("DataFmt", LTR24_FORMAT_24).toInt(&ok);
        if (ok && ((intval==LTR24_FORMAT_20) || (intval==LTR24_FORMAT_24))) {
            m_cfg.DataFmt = intval;
        } else {
            LQMeasLog->report(LogLevel::Warning, tr("Load config: invalid data format value"), this);
        }

        intval = set.value("AdcFreqCode", LTR24_FREQ_117K).toInt(&ok);
        if (ok && ((intval >= LTR24_FREQ_117K) && (intval <= LTR24_FREQ_610))) {
            cfg.ADCFreqCode = intval;
        } else {
            LQMeasLog->report(LogLevel::Warning, tr("Load config: invalid frequency code value"), this);
        }

        intval = set.value("ISrcValue", LTR24_I_SRC_VALUE_2_86).toInt(&ok);
        if (ok && ((intval==LTR24_I_SRC_VALUE_2_86) || (intval==LTR24_I_SRC_VALUE_10))) {
            m_cfg.ISrcValue = intval;
        } else {
            LQMeasLog->report(LogLevel::Warning, tr("Load config: invalid current source value"), this);
        }

        m_cfg.TestMode = set.value("TestMode").toBool();
        return adcSetRawConfig(&cfg, sizeof(cfg));
    }

    Error LTR24::protSaveConfig(QSettings &set) const {

        set.beginWriteArray("Channels", adcChannelsCnt());
        for (unsigned ch=0; ch< adcChannelsCnt(); ch++) {
            set.setArrayIndex(ch);
            set.setValue("Enabled", m_cfg.Ch[ch].Enabled);
            set.setValue("Range", m_cfg.Ch[ch].Range);
            set.setValue("ICPMode", m_cfg.Ch[ch].ICPMode);
            set.setValue("AC", m_cfg.Ch[ch].AC);
        }
        set.endArray();

        set.setValue("DataFmt", m_cfg.DataFmt);
        set.setValue("AdcFreqCode", m_cfg.ADCFreqCode);
        set.setValue("ISrcValue", m_cfg.ISrcValue);
        set.setValue("TestMode", m_cfg.TestMode);
        return Error::Success();
    }

    Error LTR24::protAdcSetRawConfig(const void *rawCfg, size_t cfg_size) {
        Error err = Error::Success();
        if (cfg_size!=sizeof(AdcConfig)) {
            err = error(LTR_ERROR_PARAMETERS);
        } else {
            memcpy(&m_cfg, rawCfg, cfg_size);
        }
        return err;
    }

    Error LTR24::rawWordsReceive(DWORD *wrds, unsigned size, unsigned tout,
                                 unsigned *recvd_size) {
        Error err = Error::Success();
        INT recv_res = LTR24_Recv(&m_hnd, wrds, 0, size, tout);
        if (recv_res >= 0) {
            *recvd_size = recv_res;
        } else {
            err = error(recv_res);
        }
        return err;
    }

    Error LTR24::protAdcStart() {
        Error err = error(LTR24_Start(&m_hnd));
        if (err.isSuccess()) {
            m_receiver.setFrameSize(adcEnabledChCnt() *
                                    (m_hnd.DataFmt == LTR24_FORMAT_20 ? 1 : 2));
        }
        return err;
    }

    Error LTR24::protAdcStop() {
        return error(LTR24_Stop(&m_hnd));
    }


    QString LTR24::serial() const {
        return QSTRING_FROM_CSTR(m_hnd.ModuleInfo.Serial);
    }

    QString LTR24::name() const {
        return typeModuleName();
    }

    QString LTR24::modificationName() const {
        return QString("%1-%2").arg(name()).arg(m_hnd.ModuleInfo.SupportICP ? "2" : "1");
    }

    int LTR24::pldVer() const {
        return m_hnd.ModuleInfo.VerPLD;
    }

    bool LTR24::supportICP() const {
        return m_hnd.ModuleInfo.SupportICP;
    }

    const double LTR24::adcRangeMaxVal(int range) const {
        return f_ranges[range];
    }

    const double LTR24::adcRangeMinVal(int range) const {
        return -f_ranges[range];
    }

    unsigned LTR24::adcRangesCnt() const {
        return sizeof(f_ranges)/sizeof(f_ranges[0]);
    }

    double LTR24::adcFreqMax() const {
        return f_adc_freqs[0];
    }

    double LTR24::typeAdcFreqVal(int idx) {
        return f_adc_freqs[idx];
    }

    unsigned LTR24::typeAdcFreqsCnt() {
     return sizeof(f_adc_freqs)/sizeof(f_adc_freqs[0]);
    }

    double LTR24::adcFreq() const {
        return m_hnd.ADCFreq;
    }

    double LTR24::adcChRange(int ch) const {
        Q_ASSERT((ch>=0) && (ch < LTR24_CHANNEL_NUM));
        return f_ranges[m_hnd.ChannelMode[ch].Range];
    }

    QString LTR24::errorString(int err) const {
        return QSTRING_FROM_CSTR(LTR24_GetErrorString(err));
    }
}
