#ifndef LQMEAS_DEVOUT_H
#define LQMEAS_DEVOUT_H

#include "DacSignals/DacSignal.h"
#include "DigSignals/DigSignal.h"
#include <QVector>
#include <QSharedPointer>
#include <QSettings>
#include "lqmeas/StdErrors.h"
#include "lqmeas/units/Units.h"

namespace LQMeas {
    class Device;

    /***********************************************************************//**
        @brief Интерфейс для вывода аналоговых и/или цифровых сигналов.

        Интерфейс предоставляет функции для однократной установки сигналов
        (циклический режим) с возможной последующей сменой сигналов.
        Для реализации потокового вывода модуль должен наследоваться
        от класса DevStreamOut (который уже наследуется от DevOut).

        Типичный вариант использования:
          - настройка модуля и сигналов (в произвольном порядке):
               - установка сигналов для нужных каналов с помощью outDacSetSignal()/outDigSetSignal()
               - разрешение интересующих каналов с помощью outDacChEnable()/outDigChEnable()
          - расчет точек и загрузка сигналов с помощью outSignalsGenSetup()
          - запуск генерации на выходах с помощь outSignalsGenStart().
          - при необходимости смены сигнала установка их с помощью  outDacSetSignal()/outDigSetSignal()
            и вызов outSignalsGenSetup()
          - при необходимости остановить генерацию вызвав outSignalsGenStop()
      ************************************************************************/
    class DevOut {
    public:
        /** Режим генерации сигнала для канала */
        enum OutChMode {
            OutChModeRam = 0, /** сигнал генерируется по точкам программным образом */
            OutChModeHard = 1 /** сигнал генерируется аппаратно по параметрам */
        };

        /** Режим работы для всех каналов с режимом #OutChModeRam */
        enum OutRamSyncMode {
            OutRamSyncModeCycle = 0,   /** Циклический режим (автогенератора) */
            OutRamSyncModeStream = 1,  /** Потоковый режим */
        };

        /* --------------- характеристики самого модуля ---------------------*/
        /** Количество каналов ЦАП в данном экземпляре модуля */
        int outDacChannelsCnt() const;
        /** Количество цифровых линий в данном экземпляре модуля */
        int outDigChannelsCnt() const;
        /** Единицы измерения величин, выдаваемых на данном канале ЦАП */
        virtual Unit   outDacUnit(int ch) const {return Units::V();}
        /** Максимальное пиковое значение для данного канала ЦАП */
        virtual double outDacMaxVal(int ch) const = 0;
        /** Минимальное пиковое значение для данного канала ЦАП */
        virtual double outDacMinVal(int ch) const = 0;
        /** Максимальное значение частоты в Гц для данного модуля */
        virtual double outDacMaxFreq() const = 0;

        /* --------------- настройка каналов и сигналов ---------------------*/

        /** Режим генерации указанного канала ЦАП */
        virtual OutChMode outDacChMode(int ch) const = 0;
        /** Режим генерации указанной цифровой линии */
        virtual OutChMode outDigChMode(int ch) const = 0;

        /** Разрешен ли указанный канал ЦАП */
        bool outDacChEnabled(int ch) const;
        /** Разрешен ли указанный канал цифровых линий */
        bool outDigChEnabled(int ch) const;

        /** Разрешение заданного канала ЦАП */
        void outDacChEnable(int ch, bool en);
        /** Запрет заданного канала ЦАП */
        void outDigChEnable(int ch, bool en);

        /** Проверка, что для указанного выхода ЦАП будет выполняться расчет
            точек для загрузки в модуль (канал разрешен, есть действительный
            сигнал и режим OutChModeRam) */
        virtual bool outDacValidRamSignal(int ch) const;
        /** Проверка, что для указанного цифрового выхода будет выполняться расчет
            точек для загрузки в модуль (канал разрешен, есть действительный
            сигнал и режим OutChModeRam) */
        virtual bool outDigValidRamSignal(int ch) const;

        /** Количество каналов для которых верна проверка outDacValidRamSignal() */
        int outDacEnabledRamChCnt() const;
        /** Количество каналов для которых верна проверка outDigValidRamSignal() */
        int outDigEnabledRamChCnt() const;


        /** Задание сигнала для указанного выхода ЦАП */
        virtual Error outDacSetSignal(int ch, QSharedPointer<DacSignal> signal);
        /** Получить установленный сигнал для указанного выхода ЦАП */
        QSharedPointer<DacSignal> outDacSignal(int ch) const;

        /** Задание сигнала для указанного цифрового выхода */
        virtual Error outDigSetSignal(int ch, QSharedPointer<DigSignal> signal);
        /** Получить установленный сигнал для указанного цифрового выхода */
        QSharedPointer<DigSignal> outDigSignal(int ch) const;

        /** Очистка всех сигналов, установленных с помощью outDacSetSignal()/
            outDigSetSignal() */
        virtual void outClearSignals();



        /** Установленная в данный момент частота вывода в Гц (на каждый канал) */
        virtual double outFreq() const = 0;
        /** Признак, запущена ли сейчас генерация данных (в любом режиме) */
        virtual bool outGenRunning() const = 0;

        /** Текущий режим работы для каналов, генерируемых по точкам из RAM */
        OutRamSyncMode outRamSyncMode() const {return m_outRamSyncMode;}
        /** Установка режима работы для каналов, генерируемых по точкам из RAM.
            Может устанавливаться автоматически в зависимости от вызываемых
            функций генерации */
        virtual Error outSetRamSyncMode(OutRamSyncMode mode) {m_outRamSyncMode = mode; return Error::Success();}


        /** Загрузка установленных сигналов и признака разрешения для
            всех каналов */
        Error outLoadSignalConfig(QSettings &set);
        /** Сохранение установленных сигналов и признака разрешения для
            всех каналов */
        Error outSaveSignalConfig(QSettings &set) const;

        /*--------- Функции для непосредственного запуска генерации ----------*/
        /** Расчет и предзагрузка точек сигнала в режиме автогенератора.
            Если генерация остановлена, то выдача сигнала начнется с
            outSignalsGenSetup(), если нет, то новый сигнал сменить
            старый (если смена на лету поддерживается) */
        Error outSignalsGenSetup();
        /** Начало генерации сигнала в режиме автогенератора. Сигнал должен
            быть уже загружен через outSignalsGenSetup() */
        Error outSignalsGenStart();
        /** Останов генерации сигнала в режиме автогенератора. */
        Error outSignalsGenStop();





        /* инициализация всех сигналов в исходное состояние (чтобы начать
           генерацию с начала сигнала) */
        void outGenRamInit();

        /* генерация следующих points точек для всех сигналов выходных
           каналов. Отсчеты ЦАП чередуются для каждого канала, для
           которого выполняется условие outDacValidRamSignal() и
           размер dac_data будет равен outDacEnabledRamChCnt() * point.
           Цифровые линии - в виде битовой маски по номеру канала (не меняется
           от количества разрешенных каналов) - поддерживается только 32 выхода
        */
        Error outGenRamData(unsigned points, QVector<double> &dac_data,
                            QVector<unsigned> &dig_data);

        /* Максимально возможный генерируемый сигнал на канал */
        virtual unsigned outGenSignalMaxSize() {return 0; }

    protected:
        DevOut(Device *dev);

        virtual Error protOutCycleLoadStart(unsigned size) = 0;
        virtual Error protOutCycleLoadFinish() = 0;
        virtual Error protOutCycleGenStart() = 0;
        virtual Error protOutCycleGenStop() = 0;
        virtual Error protOutSendData(const double *dac_data, unsigned dac_size,
                                      const unsigned *dig_data, unsigned dig_size,
                                      unsigned flags, unsigned tout) = 0;
        virtual Error protOutFlushData(unsigned tout) = 0;

        virtual unsigned outSendTout(unsigned dac_size, unsigned dout_size);


        void outSetDacChCnt(int ch_cnt);
        void outSetDigChCnt(int ch_cnt);

        Error outCalcCycleSigsParams(double out_freq, unsigned max_sig_size, unsigned *res_size);


    private:
        OutRamSyncMode m_outRamSyncMode;
        Device *m_dev;

        QVector< bool > m_dacChEnabled;
        QVector< bool > m_doutChEnabled;
        QVector< QSharedPointer<DacSignal> > m_sigsAnalog;
        QVector< QSharedPointer<DigSignal> > m_sigsDig;

        const double m_cycle_freq_rel_err;

    };
}

#endif // LQMEAS_DEVOUT_H
